@@ -24,27 +24,33 @@
 
 static void gtp_sess_timeout(ogs_gtp_xact_t *xact, void *data)
 {
-    sgwc_sess_t *sess = data;
+    sgwc_sess_t *sess = NULL;
+    ogs_pool_id_t sess_id = OGS_INVALID_POOL_ID;
     sgwc_ue_t *sgwc_ue = NULL;
     uint8_t type = 0;
 
     ogs_assert(xact);
-    ogs_assert(sess);
-    sgwc_ue = sess->sgwc_ue;
-    ogs_assert(sgwc_ue);
-
     type = xact->seq[0].type;
 
+    ogs_assert(data);
+    sess_id = OGS_POINTER_TO_UINT(data);
+    ogs_assert(sess_id >= OGS_MIN_POOL_ID && sess_id <= OGS_MAX_POOL_ID);
+
+    sess = sgwc_sess_find_by_id(sess_id);
+    if (!sess) {
+        ogs_error("Session has already been removed [%d]", type);
+        return;
+    }
+
+    sgwc_ue = sgwc_ue_find_by_id(sess->sgwc_ue_id);
+    ogs_assert(sgwc_ue);
+
     switch (type) {
     case OGS_GTP2_DELETE_SESSION_REQUEST_TYPE:
         ogs_error("[%s] No Delete Session Response", sgwc_ue->imsi_bcd);
-        if (!sgwc_sess_cycle(sess)) {
-            ogs_error("[%s] Session has already been removed",
-                    sgwc_ue->imsi_bcd);
-            break;
-        }
         ogs_assert(OGS_OK ==
-            sgwc_pfcp_send_session_deletion_request(sess, NULL, NULL));
+            sgwc_pfcp_send_session_deletion_request(
+                sess, OGS_INVALID_POOL_ID, NULL));
         break;
     default:
         ogs_error("GTP Timeout : IMSI[%s] Message-Type[%d]",
@@ -54,31 +60,53 @@ static void gtp_sess_timeout(ogs_gtp_xact_t *xact, void *data)
 
 static void gtp_bearer_timeout(ogs_gtp_xact_t *xact, void *data)
 {
-    sgwc_bearer_t *bearer = data;
+    sgwc_bearer_t *bearer = NULL;
+    ogs_pool_id_t bearer_id = OGS_INVALID_POOL_ID;
     sgwc_sess_t *sess = NULL;
     sgwc_ue_t *sgwc_ue = NULL;
     uint8_t type = 0;
 
     ogs_assert(xact);
-    ogs_assert(bearer);
-    sess = bearer->sess;
-    ogs_assert(sess);
-    sgwc_ue = sess->sgwc_ue;
-    ogs_assert(sgwc_ue);
-
     type = xact->seq[0].type;
 
+    ogs_assert(data);
+    bearer_id = OGS_POINTER_TO_UINT(data);
+    ogs_assert(bearer_id >= OGS_MIN_POOL_ID && bearer_id <= OGS_MAX_POOL_ID);
+
+    bearer = sgwc_bearer_find_by_id(bearer_id);
+    if (!bearer) {
+        ogs_error("Bearer has already been removed [%d]", type);
+        return;
+    }
+
+    sess = sgwc_sess_find_by_id(bearer->sess_id);
+    ogs_assert(sess);
+    sgwc_ue = sgwc_ue_find_by_id(sess->sgwc_ue_id);
+    ogs_assert(sgwc_ue);
+
     ogs_error("GTP Timeout : IMSI[%s] Message-Type[%d]",
             sgwc_ue->imsi_bcd, type);
 }
 
 static void pfcp_sess_timeout(ogs_pfcp_xact_t *xact, void *data)
 {
+    sgwc_sess_t *sess = NULL;
+    ogs_pool_id_t sess_id = OGS_INVALID_POOL_ID;
     uint8_t type;
 
     ogs_assert(xact);
     type = xact->seq[0].type;
 
+    ogs_assert(data);
+    sess_id = OGS_POINTER_TO_UINT(data);
+    ogs_assert(sess_id >= OGS_MIN_POOL_ID && sess_id <= OGS_MAX_POOL_ID);
+
+    sess = sgwc_sess_find_by_id(sess_id);
+    if (!sess) {
+        ogs_error("Session has already been removed [%d]", type);
+        return;
+    }
+
     switch (type) {
     case OGS_PFCP_SESSION_ESTABLISHMENT_REQUEST_TYPE:
         ogs_error("No PFCP session establishment response");
@@ -203,6 +231,12 @@ void sgwc_s11_handle_create_session_request(
     if (req->access_point_name.presence == 0) {
         ogs_error("No APN");
         cause_value = OGS_GTP2_CAUSE_MANDATORY_IE_MISSING;
+    } else {
+        if (ogs_fqdn_parse(apn, req->access_point_name.data,
+            ogs_min(req->access_point_name.len, OGS_MAX_APN_LEN)) <= 0) {
+            ogs_error("Invalid APN");
+            cause_value = OGS_GTP2_CAUSE_MANDATORY_IE_INCORRECT;
+        }
     }
     if (req->sender_f_teid_for_control_plane.presence == 0) {
         ogs_error("No Sender F-TEID");
@@ -221,9 +255,6 @@ void sgwc_s11_handle_create_session_request(
     }
 
     /* Add Session */
-    ogs_assert(0 < ogs_fqdn_parse(apn,
-            req->access_point_name.data,
-            ogs_min(req->access_point_name.len, OGS_MAX_APN_LEN)));
     sess = sgwc_sess_find_by_ebi(sgwc_ue,
             req->bearer_contexts_to_be_created[0].eps_bearer_id.u8);
     if (sess) {
@@ -239,21 +270,22 @@ void sgwc_s11_handle_create_session_request(
     /* Set User Location Information */
     if (req->user_location_information.presence == 1) {
         decoded = ogs_gtp2_parse_uli(&uli, &req->user_location_information);
-        ogs_assert(req->user_location_information.len == decoded);
+        if (req->user_location_information.len == decoded) {
+            sgwc_ue->uli_presence = true;
 
-        sgwc_ue->uli_presence = true;
+            ogs_nas_to_plmn_id(&sgwc_ue->e_tai.plmn_id, &uli.tai.nas_plmn_id);
+            sgwc_ue->e_tai.tac = uli.tai.tac;
+            ogs_nas_to_plmn_id(&sgwc_ue->e_cgi.plmn_id, &uli.e_cgi.nas_plmn_id);
+            sgwc_ue->e_cgi.cell_id = uli.e_cgi.cell_id;
 
-        ogs_nas_to_plmn_id(&sgwc_ue->e_tai.plmn_id, &uli.tai.nas_plmn_id);
-        sgwc_ue->e_tai.tac = uli.tai.tac;
-        ogs_nas_to_plmn_id(&sgwc_ue->e_cgi.plmn_id, &uli.e_cgi.nas_plmn_id);
-        sgwc_ue->e_cgi.cell_id = uli.e_cgi.cell_id;
-
-        ogs_debug("    TAI[PLMN_ID:%06x,TAC:%d]",
-                ogs_plmn_id_hexdump(&sgwc_ue->e_tai.plmn_id),
-                sgwc_ue->e_tai.tac);
-        ogs_debug("    E_CGI[PLMN_ID:%06x,CELL_ID:0x%x]",
-                ogs_plmn_id_hexdump(&sgwc_ue->e_cgi.plmn_id),
-                sgwc_ue->e_cgi.cell_id);
+            ogs_debug("    TAI[PLMN_ID:%06x,TAC:%d]",
+                    ogs_plmn_id_hexdump(&sgwc_ue->e_tai.plmn_id),
+                    sgwc_ue->e_tai.tac);
+            ogs_debug("    E_CGI[PLMN_ID:%06x,CELL_ID:0x%x]",
+                    ogs_plmn_id_hexdump(&sgwc_ue->e_cgi.plmn_id),
+                    sgwc_ue->e_cgi.cell_id);
+        } else
+            ogs_error("Invalid User Location Info(ULI)");
     }
 
     /* Select SGW-U based on UE Location Information */
@@ -379,7 +411,7 @@ void sgwc_s11_handle_create_session_request(
 
     ogs_assert(OGS_OK ==
         sgwc_pfcp_send_session_establishment_request(
-            sess, s11_xact, gtpbuf, 0));
+            sess, s11_xact->id, gtpbuf, 0));
 }
 
 void sgwc_s11_handle_modify_bearer_request(
@@ -396,6 +428,7 @@ void sgwc_s11_handle_modify_bearer_request(
     sgwc_sess_t *sess = NULL;
     sgwc_bearer_t *bearer = NULL;
     sgwc_tunnel_t *dl_tunnel = NULL;
+    ogs_pfcp_pdr_t *pdr = NULL;
     ogs_pfcp_far_t *far = NULL;
     ogs_ip_t remote_ip;
     ogs_ip_t zero_ip;
@@ -461,24 +494,32 @@ void sgwc_s11_handle_modify_bearer_request(
             break;
         }
 
-        sess = bearer->sess;
+        sess = sgwc_sess_find_by_id(bearer->sess_id);
         ogs_assert(sess);
 
         ogs_list_for_each_entry(&pfcp_xact_list, pfcp_xact, tmpnode) {
-            if (sess == pfcp_xact->data) {
-                current_xact = pfcp_xact;
-                break;
+            if (pfcp_xact->modify_flags & OGS_PFCP_MODIFY_SESSION) {
+                ogs_pool_id_t sess_id = OGS_POINTER_TO_UINT(pfcp_xact->data);
+                ogs_assert(sess_id >= OGS_MIN_POOL_ID &&
+                        sess_id <= OGS_MAX_POOL_ID);
+                if (sess->id == sess_id) {
+                    current_xact = pfcp_xact;
+                    break;
+                }
             }
         }
 
         if (!current_xact) {
             current_xact = ogs_pfcp_xact_local_create(
-                    sess->pfcp_node, pfcp_sess_timeout, sess);
+                    sess->pfcp_node, pfcp_sess_timeout,
+                    OGS_UINT_TO_POINTER(sess->id));
             ogs_assert(current_xact);
 
-            current_xact->assoc_xact = s11_xact;
-            current_xact->modify_flags = OGS_PFCP_MODIFY_SESSION|
-                OGS_PFCP_MODIFY_DL_ONLY|OGS_PFCP_MODIFY_ACTIVATE;
+            current_xact->assoc_xact_id = s11_xact->id;
+            current_xact->modify_flags =
+                OGS_PFCP_MODIFY_SESSION|OGS_PFCP_MODIFY_DL_ONLY|
+                OGS_PFCP_MODIFY_OUTER_HEADER_REMOVAL|
+                OGS_PFCP_MODIFY_ACTIVATE;
             if (gtpbuf) {
                 current_xact->gtpbuf = ogs_pkbuf_copy(gtpbuf);
                 ogs_assert(current_xact->gtpbuf);
@@ -515,6 +556,13 @@ void sgwc_s11_handle_modify_bearer_request(
 
         memcpy(&dl_tunnel->remote_ip, &remote_ip, sizeof(ogs_ip_t));
 
+        pdr = dl_tunnel->pdr;
+        ogs_assert(pdr);
+
+        pdr->outer_header_removal_len = 1;
+        pdr->outer_header_removal.description =
+            OGS_PFCP_OUTER_HEADER_REMOVAL_GTPU_UDP_IP;
+
         far = dl_tunnel->far;
         ogs_assert(far);
 
@@ -539,21 +587,22 @@ void sgwc_s11_handle_modify_bearer_request(
 
     if (req->user_location_information.presence == 1) {
         decoded = ogs_gtp2_parse_uli(&uli, &req->user_location_information);
-        ogs_assert(req->user_location_information.len == decoded);
+        if (req->user_location_information.len == decoded) {
+            sgwc_ue->uli_presence = true;
 
-        sgwc_ue->uli_presence = true;
+            ogs_nas_to_plmn_id(&sgwc_ue->e_tai.plmn_id, &uli.tai.nas_plmn_id);
+            sgwc_ue->e_tai.tac = uli.tai.tac;
+            ogs_nas_to_plmn_id(&sgwc_ue->e_cgi.plmn_id, &uli.e_cgi.nas_plmn_id);
+            sgwc_ue->e_cgi.cell_id = uli.e_cgi.cell_id;
 
-        ogs_nas_to_plmn_id(&sgwc_ue->e_tai.plmn_id, &uli.tai.nas_plmn_id);
-        sgwc_ue->e_tai.tac = uli.tai.tac;
-        ogs_nas_to_plmn_id(&sgwc_ue->e_cgi.plmn_id, &uli.e_cgi.nas_plmn_id);
-        sgwc_ue->e_cgi.cell_id = uli.e_cgi.cell_id;
-
-        ogs_debug("    TAI[PLMN_ID:%06x,TAC:%d]",
-                ogs_plmn_id_hexdump(&sgwc_ue->e_tai.plmn_id),
-                sgwc_ue->e_tai.tac);
-        ogs_debug("    E_CGI[PLMN_ID:%06x,CELL_ID:0x%x]",
-                ogs_plmn_id_hexdump(&sgwc_ue->e_cgi.plmn_id),
-                sgwc_ue->e_cgi.cell_id);
+            ogs_debug("    TAI[PLMN_ID:%06x,TAC:%d]",
+                    ogs_plmn_id_hexdump(&sgwc_ue->e_tai.plmn_id),
+                    sgwc_ue->e_tai.tac);
+            ogs_debug("    E_CGI[PLMN_ID:%06x,CELL_ID:0x%x]",
+                    ogs_plmn_id_hexdump(&sgwc_ue->e_cgi.plmn_id),
+                    sgwc_ue->e_cgi.cell_id);
+        } else
+            ogs_error("Invalid User Location Info(ULI)");
     }
 
     ogs_debug("    MME_S11_TEID[%d] SGW_S11_TEID[%d]",
@@ -561,8 +610,20 @@ void sgwc_s11_handle_modify_bearer_request(
     ogs_debug("    ENB_S1U_TEID[%d] SGW_S1U_TEID[%d]",
         dl_tunnel->remote_teid, dl_tunnel->local_teid);
 
-    ogs_list_for_each_entry(&pfcp_xact_list, pfcp_xact, tmpnode)
-        sgwc_pfcp_send_bearer_to_modify_list(pfcp_xact->data, pfcp_xact);
+    ogs_list_for_each_entry(&pfcp_xact_list, pfcp_xact, tmpnode) {
+        if (pfcp_xact->modify_flags & OGS_PFCP_MODIFY_SESSION) {
+            sgwc_sess_t *sess = NULL;
+
+            ogs_pool_id_t sess_id = OGS_POINTER_TO_UINT(pfcp_xact->data);
+            ogs_assert(sess_id >= OGS_MIN_POOL_ID &&
+                    sess_id <= OGS_MAX_POOL_ID);
+
+            sess = sgwc_sess_find_by_id(sess_id);
+            ogs_assert(sess);
+
+            sgwc_pfcp_send_bearer_to_modify_list(sess, pfcp_xact);
+        }
+    }
 }
 
 void sgwc_s11_handle_delete_session_request(
@@ -652,7 +713,8 @@ void sgwc_s11_handle_delete_session_request(
         indication->scope_indication == 1) {
 
         ogs_assert(OGS_OK ==
-            sgwc_pfcp_send_session_deletion_request(sess, s11_xact, gtpbuf));
+            sgwc_pfcp_send_session_deletion_request(
+                sess, s11_xact->id, gtpbuf));
 
     } else {
         message->h.type = OGS_GTP2_DELETE_SESSION_REQUEST_TYPE;
@@ -665,7 +727,8 @@ void sgwc_s11_handle_delete_session_request(
         }
 
         s5c_xact = ogs_gtp_xact_local_create(
-                sess->gnode, &message->h, gtpbuf, gtp_sess_timeout, sess);
+                sess->gnode, &message->h, gtpbuf, gtp_sess_timeout,
+                OGS_UINT_TO_POINTER(sess->id));
         if (!s5c_xact) {
             ogs_error("ogs_gtp_xact_local_create() failed");
             return;
@@ -690,7 +753,9 @@ void sgwc_s11_handle_create_bearer_response(
 
     sgwc_sess_t *sess = NULL;
     sgwc_bearer_t *bearer = NULL;
+    ogs_pool_id_t bearer_id = OGS_INVALID_POOL_ID;
     sgwc_tunnel_t *dl_tunnel = NULL, *ul_tunnel = NULL;
+    ogs_pfcp_pdr_t *pdr = NULL;
     ogs_pfcp_far_t *far = NULL;
 
     ogs_gtp_xact_t *s5c_xact = NULL;
@@ -710,18 +775,35 @@ void sgwc_s11_handle_create_bearer_response(
      * Check Transaction
      ********************/
     ogs_assert(s11_xact);
-    s5c_xact = s11_xact->assoc_xact;
+    s5c_xact = ogs_gtp_xact_find_by_id(s11_xact->assoc_xact_id);
     ogs_assert(s5c_xact);
 
-    if (s11_xact->xid & OGS_GTP_CMD_XACT_ID)
+    if (s11_xact->xid & OGS_GTP_CMD_XACT_ID) {
         /* MME received Bearer Resource Modification Request */
-        bearer = s5c_xact->data;
-    else
-        bearer = s11_xact->data;
+        ogs_assert(s5c_xact->data);
+        bearer_id = OGS_POINTER_TO_UINT(s5c_xact->data);
+        ogs_assert(bearer_id >= OGS_MIN_POOL_ID &&
+                bearer_id <= OGS_MAX_POOL_ID);
 
-    ogs_assert(bearer);
-    sess = bearer->sess;
-    ogs_assert(sess);
+        bearer = sgwc_bearer_find_by_id(bearer_id);
+        if (!bearer)
+            ogs_error("No Bearer ID [%d]", bearer_id);
+    } else {
+        ogs_assert(s11_xact->data);
+        bearer_id = OGS_POINTER_TO_UINT(s11_xact->data);
+        ogs_assert(bearer_id >= OGS_MIN_POOL_ID &&
+                bearer_id <= OGS_MAX_POOL_ID);
+
+        bearer = sgwc_bearer_find_by_id(bearer_id);
+        if (!bearer)
+            ogs_error("No Bearer ID [%d]", bearer_id);
+    }
+
+    if (bearer) {
+        sess = sgwc_sess_find_by_id(bearer->sess_id);
+        if (!sess)
+            ogs_error("No Session ID [%d]", bearer->sess_id);
+    }
 
     rv = ogs_gtp_xact_commit(s11_xact);
     ogs_expect(rv == OGS_OK);
@@ -757,11 +839,22 @@ void sgwc_s11_handle_create_bearer_response(
         cause_value = OGS_GTP2_CAUSE_MANDATORY_IE_MISSING;
     }
 
+    if (!bearer) {
+        ogs_error("No Bearer Context");
+        cause_value = OGS_GTP2_CAUSE_CONTEXT_NOT_FOUND;
+    }
+    if (!sess) {
+        ogs_error("No Session Context");
+        cause_value = OGS_GTP2_CAUSE_CONTEXT_NOT_FOUND;
+    }
+
     if (cause_value != OGS_GTP2_CAUSE_REQUEST_ACCEPTED) {
-        ogs_assert(OGS_OK ==
-            sgwc_pfcp_send_bearer_modification_request(
-                bearer, NULL, NULL,
-                OGS_PFCP_MODIFY_UL_ONLY|OGS_PFCP_MODIFY_REMOVE));
+        if (bearer) {
+            ogs_assert(OGS_OK ==
+                sgwc_pfcp_send_bearer_modification_request(
+                    bearer, OGS_INVALID_POOL_ID, NULL,
+                    OGS_PFCP_MODIFY_UL_ONLY|OGS_PFCP_MODIFY_REMOVE));
+        }
         ogs_gtp_send_error_message(s5c_xact, sess ? sess->pgw_s5c_teid : 0,
                 OGS_GTP2_CREATE_BEARER_RESPONSE_TYPE, cause_value);
         return;
@@ -779,7 +872,7 @@ void sgwc_s11_handle_create_bearer_response(
         ogs_error("GTP Cause [Value:%d]", cause_value);
         ogs_assert(OGS_OK ==
             sgwc_pfcp_send_bearer_modification_request(
-                bearer, NULL, NULL,
+                bearer, OGS_INVALID_POOL_ID, NULL,
                 OGS_PFCP_MODIFY_UL_ONLY|OGS_PFCP_MODIFY_REMOVE));
         ogs_gtp_send_error_message(s5c_xact, sess ? sess->pgw_s5c_teid : 0,
                 OGS_GTP2_CREATE_BEARER_RESPONSE_TYPE, cause_value);
@@ -799,9 +892,23 @@ void sgwc_s11_handle_create_bearer_response(
 
     /* Find the Tunnel by SGW-S1U-TEID */
     ul_tunnel = sgwc_tunnel_find_by_teid(sgwc_ue, be32toh(sgw_s1u_teid->teid));
-    ogs_assert(ul_tunnel);
+    if (!ul_tunnel) {
+        ogs_error("No UL-tunnel [EBI:%d, TEID:0x%x]",
+                bearer->ebi, be32toh(sgw_s1u_teid->teid));
+        ogs_gtp_send_error_message(s5c_xact, sess ? sess->pgw_s5c_teid : 0,
+                OGS_GTP2_CREATE_BEARER_RESPONSE_TYPE,
+                OGS_GTP2_CAUSE_GRE_KEY_NOT_FOUND);
+        return;
+    }
     dl_tunnel = sgwc_dl_tunnel_in_bearer(bearer);
-    ogs_assert(dl_tunnel);
+    if (!dl_tunnel) {
+        ogs_error("No DL-tunnel [EBI:%d, TEID:0x%x]",
+                bearer->ebi, be32toh(sgw_s1u_teid->teid));
+        ogs_gtp_send_error_message(s5c_xact, sess ? sess->pgw_s5c_teid : 0,
+                OGS_GTP2_CREATE_BEARER_RESPONSE_TYPE,
+                OGS_GTP2_CAUSE_CONTEXT_NOT_FOUND);
+        return;
+    }
 
     /* Set EBI */
     bearer->ebi = rsp->bearer_contexts.eps_bearer_id.u8;
@@ -816,6 +923,13 @@ void sgwc_s11_handle_create_bearer_response(
     ogs_assert(OGS_OK ==
             ogs_gtp2_f_teid_to_ip(enb_s1u_teid, &dl_tunnel->remote_ip));
 
+    pdr = dl_tunnel->pdr;
+    ogs_assert(pdr);
+
+    pdr->outer_header_removal_len = 1;
+    pdr->outer_header_removal.description =
+        OGS_PFCP_OUTER_HEADER_REMOVAL_GTPU_UDP_IP;
+
     far = dl_tunnel->far;
     ogs_assert(far);
 
@@ -828,26 +942,27 @@ void sgwc_s11_handle_create_bearer_response(
 
     if (rsp->user_location_information.presence == 1) {
         decoded = ogs_gtp2_parse_uli(&uli, &rsp->user_location_information);
-        ogs_assert(rsp->user_location_information.len == decoded);
+        if (rsp->user_location_information.len == decoded) {
+            sgwc_ue->uli_presence = true;
 
-        sgwc_ue->uli_presence = true;
+            ogs_nas_to_plmn_id(&sgwc_ue->e_tai.plmn_id, &uli.tai.nas_plmn_id);
+            sgwc_ue->e_tai.tac = uli.tai.tac;
+            ogs_nas_to_plmn_id(&sgwc_ue->e_cgi.plmn_id, &uli.e_cgi.nas_plmn_id);
+            sgwc_ue->e_cgi.cell_id = uli.e_cgi.cell_id;
 
-        ogs_nas_to_plmn_id(&sgwc_ue->e_tai.plmn_id, &uli.tai.nas_plmn_id);
-        sgwc_ue->e_tai.tac = uli.tai.tac;
-        ogs_nas_to_plmn_id(&sgwc_ue->e_cgi.plmn_id, &uli.e_cgi.nas_plmn_id);
-        sgwc_ue->e_cgi.cell_id = uli.e_cgi.cell_id;
-
-        ogs_debug("    TAI[PLMN_ID:%06x,TAC:%d]",
-                ogs_plmn_id_hexdump(&sgwc_ue->e_tai.plmn_id),
-                sgwc_ue->e_tai.tac);
-        ogs_debug("    E_CGI[PLMN_ID:%06x,CELL_ID:0x%x]",
-                ogs_plmn_id_hexdump(&sgwc_ue->e_cgi.plmn_id),
-                sgwc_ue->e_cgi.cell_id);
+            ogs_debug("    TAI[PLMN_ID:%06x,TAC:%d]",
+                    ogs_plmn_id_hexdump(&sgwc_ue->e_tai.plmn_id),
+                    sgwc_ue->e_tai.tac);
+            ogs_debug("    E_CGI[PLMN_ID:%06x,CELL_ID:0x%x]",
+                    ogs_plmn_id_hexdump(&sgwc_ue->e_cgi.plmn_id),
+                    sgwc_ue->e_cgi.cell_id);
+        } else
+            ogs_error("Invalid User Location Info(ULI)");
     }
 
     ogs_assert(OGS_OK ==
         sgwc_pfcp_send_bearer_modification_request(
-            bearer, s5c_xact, gtpbuf,
+            bearer, s5c_xact->id, gtpbuf,
             OGS_PFCP_MODIFY_DL_ONLY|OGS_PFCP_MODIFY_CREATE));
 }
 
@@ -862,6 +977,7 @@ void sgwc_s11_handle_update_bearer_response(
     ogs_gtp_xact_t *s5c_xact = NULL;
     sgwc_sess_t *sess = NULL;
     sgwc_bearer_t *bearer = NULL;
+    ogs_pool_id_t bearer_id = OGS_INVALID_POOL_ID;
     ogs_gtp2_update_bearer_response_t *rsp = NULL;
 
     ogs_assert(sgwc_ue);
@@ -875,17 +991,29 @@ void sgwc_s11_handle_update_bearer_response(
      * Check Transaction
      ********************/
     ogs_assert(s11_xact);
-    s5c_xact = s11_xact->assoc_xact;
+    s5c_xact = ogs_gtp_xact_find_by_id(s11_xact->assoc_xact_id);
     ogs_assert(s5c_xact);
 
-    if (s11_xact->xid & OGS_GTP_CMD_XACT_ID)
+    if (s11_xact->xid & OGS_GTP_CMD_XACT_ID) {
         /* MME received Bearer Resource Modification Request */
-        bearer = s5c_xact->data;
-    else
-        bearer = s11_xact->data;
+        ogs_assert(s5c_xact->data);
+        bearer_id = OGS_POINTER_TO_UINT(s5c_xact->data);
+        ogs_assert(bearer_id >= OGS_MIN_POOL_ID &&
+                bearer_id <= OGS_MAX_POOL_ID);
 
-    ogs_assert(bearer);
-    sess = bearer->sess;
+        bearer = sgwc_bearer_find_by_id(bearer_id);
+        ogs_assert(bearer);
+    } else {
+        ogs_assert(s11_xact->data);
+        bearer_id = OGS_POINTER_TO_UINT(s11_xact->data);
+        ogs_assert(bearer_id >= OGS_MIN_POOL_ID &&
+                bearer_id <= OGS_MAX_POOL_ID);
+
+        bearer = sgwc_bearer_find_by_id(bearer_id);
+        ogs_assert(bearer);
+    }
+
+    sess = sgwc_sess_find_by_id(bearer->sess_id);
     ogs_assert(sess);
 
     rv = ogs_gtp_xact_commit(s11_xact);
@@ -985,6 +1113,7 @@ void sgwc_s11_handle_delete_bearer_response(
 
     sgwc_sess_t *sess = NULL;
     sgwc_bearer_t *bearer = NULL;
+    ogs_pool_id_t bearer_id = OGS_INVALID_POOL_ID;
     ogs_gtp2_delete_bearer_response_t *rsp = NULL;
 
     ogs_assert(sgwc_ue);
@@ -998,22 +1127,43 @@ void sgwc_s11_handle_delete_bearer_response(
      * Check Transaction
      ********************/
     ogs_assert(s11_xact);
-    s5c_xact = s11_xact->assoc_xact;
+    s5c_xact = ogs_gtp_xact_find_by_id(s11_xact->assoc_xact_id);
     ogs_assert(s5c_xact);
 
-    if (s11_xact->xid & OGS_GTP_CMD_XACT_ID)
+    if (s11_xact->xid & OGS_GTP_CMD_XACT_ID) {
         /* MME received Bearer Resource Modification Request */
-        bearer = s5c_xact->data;
-    else
-        bearer = s11_xact->data;
+        ogs_assert(s5c_xact->data);
+        bearer_id = OGS_POINTER_TO_UINT(s5c_xact->data);
+        ogs_assert(bearer_id >= OGS_MIN_POOL_ID &&
+                bearer_id <= OGS_MAX_POOL_ID);
 
-    ogs_assert(bearer);
-    sess = bearer->sess;
+        rv = ogs_gtp_xact_commit(s11_xact);
+        ogs_expect(rv == OGS_OK);
+
+        bearer = sgwc_bearer_find_by_id(bearer_id);
+        if (!bearer) {
+            ogs_error("No Bearer Context [%d]", bearer_id);
+            return;
+        }
+    } else {
+        ogs_assert(s11_xact->data);
+        bearer_id = OGS_POINTER_TO_UINT(s11_xact->data);
+        ogs_assert(bearer_id >= OGS_MIN_POOL_ID &&
+                bearer_id <= OGS_MAX_POOL_ID);
+
+        rv = ogs_gtp_xact_commit(s11_xact);
+        ogs_expect(rv == OGS_OK);
+
+        bearer = sgwc_bearer_find_by_id(bearer_id);
+        if (!bearer) {
+            ogs_error("No Bearer Context [ID:%d]", bearer_id);
+            return;
+        }
+    }
+
+    sess = sgwc_sess_find_by_id(bearer->sess_id);
     ogs_assert(sess);
 
-    rv = ogs_gtp_xact_commit(s11_xact);
-    ogs_expect(rv == OGS_OK);
-
     /************************
      * Check SGWC-UE Context
      ************************/
@@ -1045,7 +1195,8 @@ void sgwc_s11_handle_delete_bearer_response(
         }
 
         ogs_assert(OGS_OK ==
-            sgwc_pfcp_send_session_deletion_request(sess, s5c_xact, gtpbuf));
+            sgwc_pfcp_send_session_deletion_request(
+                sess, s5c_xact->id, gtpbuf));
     } else {
        /*
         * << EPS Bearer IDs >>
@@ -1097,7 +1248,7 @@ void sgwc_s11_handle_delete_bearer_response(
 
         ogs_assert(OGS_OK ==
             sgwc_pfcp_send_bearer_modification_request(
-                bearer, s5c_xact, gtpbuf, OGS_PFCP_MODIFY_REMOVE));
+                bearer, s5c_xact->id, gtpbuf, OGS_PFCP_MODIFY_REMOVE));
     }
 }
 
@@ -1144,9 +1295,10 @@ void sgwc_s11_handle_release_access_bearers_request(
 
     ogs_list_for_each(&sgwc_ue->sess_list, sess) {
 
+        ogs_assert(ogs_list_count(&sess->bearer_list));
         ogs_assert(OGS_OK ==
             sgwc_pfcp_send_session_modification_request(
-                sess, s11_xact, gtpbuf,
+                sess, s11_xact->id, gtpbuf,
                 OGS_PFCP_MODIFY_DL_ONLY|OGS_PFCP_MODIFY_DEACTIVATE));
     }
 }
@@ -1159,6 +1311,7 @@ void sgwc_s11_handle_downlink_data_notification_ack(
     uint8_t cause_value;
 
     sgwc_bearer_t *bearer = NULL;
+    ogs_pool_id_t bearer_id = OGS_INVALID_POOL_ID;
     sgwc_sess_t *sess = NULL;
 
     ogs_gtp2_downlink_data_notification_acknowledge_t *ack = NULL;
@@ -1171,9 +1324,14 @@ void sgwc_s11_handle_downlink_data_notification_ack(
      * Check Transaction
      ********************/
     ogs_assert(s11_xact);
-    bearer = s11_xact->data;
+    ogs_assert(s11_xact->data);
+    bearer_id = OGS_POINTER_TO_UINT(s11_xact->data);
+    ogs_assert(bearer_id >= OGS_MIN_POOL_ID &&
+            bearer_id <= OGS_MAX_POOL_ID);
+
+    bearer = sgwc_bearer_find_by_id(bearer_id);
     ogs_assert(bearer);
-    sess = bearer->sess;
+    sess = sgwc_sess_find_by_id(bearer->sess_id);
     ogs_assert(sess);
 
     rv = ogs_gtp_xact_commit(s11_xact);
@@ -1195,7 +1353,7 @@ void sgwc_s11_handle_downlink_data_notification_ack(
         ogs_error("No Cause");
     }
 
-    ogs_debug("Downlink Data Notification Acknowledge");
+    ogs_debug("Downlink Data Notification Acknowledge [%d]", bearer->id);
     if (sgwc_ue) {
         ogs_debug("    MME_S11_TEID[%d] SGW_S11_TEID[%d]",
             sgwc_ue->mme_s11_teid, sgwc_ue->sgw_s11_teid);
@@ -1211,6 +1369,7 @@ void sgwc_s11_handle_create_indirect_data_forwarding_tunnel_request(
     sgwc_sess_t *sess = NULL;
     sgwc_bearer_t *bearer = NULL;
     sgwc_tunnel_t *tunnel = NULL;
+    ogs_pfcp_pdr_t *pdr = NULL;
     ogs_pfcp_far_t *far = NULL;
 
     ogs_gtp2_create_indirect_data_forwarding_tunnel_request_t *req = NULL;
@@ -1262,7 +1421,18 @@ void sgwc_s11_handle_create_indirect_data_forwarding_tunnel_request(
 
         bearer = sgwc_bearer_find_by_ue_ebi(sgwc_ue,
                     req->bearer_contexts[i].eps_bearer_id.u8);
-        ogs_assert(bearer);
+        if (!bearer) {
+            ogs_error("No Bearer Context [%d]",
+                    req->bearer_contexts[i].eps_bearer_id.u8);
+            ogs_gtp_send_error_message(
+                s11_xact, sgwc_ue ? sgwc_ue->mme_s11_teid : 0,
+                OGS_GTP2_CREATE_INDIRECT_DATA_FORWARDING_TUNNEL_RESPONSE_TYPE,
+                OGS_GTP2_CAUSE_CONTEXT_NOT_FOUND);
+            return;
+        }
+
+        sess = sgwc_sess_find_by_id(bearer->sess_id);
+        ogs_assert(sess);
 
         if (req->bearer_contexts[i].s1_u_enodeb_f_teid.presence) {
             req_teid = req->bearer_contexts[i].s1_u_enodeb_f_teid.data;
@@ -1270,7 +1440,14 @@ void sgwc_s11_handle_create_indirect_data_forwarding_tunnel_request(
 
             tunnel = sgwc_tunnel_add(bearer,
                     OGS_GTP2_F_TEID_SGW_GTP_U_FOR_DL_DATA_FORWARDING);
-            ogs_assert(tunnel);
+            if (!tunnel) {
+                ogs_error("sgwc_tunnel_add() failed");
+                ogs_gtp_send_error_message(
+                    s11_xact, sgwc_ue ? sgwc_ue->mme_s11_teid : 0,
+                    OGS_GTP2_CREATE_INDIRECT_DATA_FORWARDING_TUNNEL_RESPONSE_TYPE,
+                    OGS_GTP2_CAUSE_SYSTEM_FAILURE);
+                return;
+            }
 
             tunnel->remote_teid = be32toh(req_teid->teid);
 
@@ -1283,6 +1460,13 @@ void sgwc_s11_handle_create_indirect_data_forwarding_tunnel_request(
                 return;
             }
 
+            pdr = tunnel->pdr;
+            ogs_assert(pdr);
+
+            pdr->outer_header_removal_len = 1;
+            pdr->outer_header_removal.description =
+                OGS_PFCP_OUTER_HEADER_REMOVAL_GTPU_UDP_IP;
+
             far = tunnel->far;
             ogs_assert(far);
 
@@ -1305,7 +1489,14 @@ void sgwc_s11_handle_create_indirect_data_forwarding_tunnel_request(
 
             tunnel = sgwc_tunnel_add(bearer,
                     OGS_GTP2_F_TEID_SGW_GTP_U_FOR_UL_DATA_FORWARDING);
-            ogs_assert(tunnel);
+            if (!tunnel) {
+                ogs_error("sgwc_tunnel_add() failed");
+                ogs_gtp_send_error_message(
+                    s11_xact, sgwc_ue ? sgwc_ue->mme_s11_teid : 0,
+                    OGS_GTP2_CREATE_INDIRECT_DATA_FORWARDING_TUNNEL_RESPONSE_TYPE,
+                    OGS_GTP2_CAUSE_SYSTEM_FAILURE);
+                return;
+            }
 
             tunnel->remote_teid = be32toh(req_teid->teid);
 
@@ -1318,6 +1509,13 @@ void sgwc_s11_handle_create_indirect_data_forwarding_tunnel_request(
                 return;
             }
 
+            pdr = tunnel->pdr;
+            ogs_assert(pdr);
+
+            pdr->outer_header_removal_len = 1;
+            pdr->outer_header_removal.description =
+                OGS_PFCP_OUTER_HEADER_REMOVAL_GTPU_UDP_IP;
+
             far = tunnel->far;
             ogs_assert(far);
 
@@ -1337,9 +1535,10 @@ void sgwc_s11_handle_create_indirect_data_forwarding_tunnel_request(
 
     ogs_list_for_each(&sgwc_ue->sess_list, sess) {
 
+        ogs_assert(ogs_list_count(&sess->bearer_list));
         ogs_assert(OGS_OK ==
             sgwc_pfcp_send_session_modification_request(
-                sess, s11_xact, gtpbuf,
+                sess, s11_xact->id, gtpbuf,
                 OGS_PFCP_MODIFY_INDIRECT|OGS_PFCP_MODIFY_CREATE));
     }
 }
@@ -1349,6 +1548,8 @@ void sgwc_s11_handle_delete_indirect_data_forwarding_tunnel_request(
         ogs_pkbuf_t *gtpbuf, ogs_gtp2_message_t *recv_message)
 {
     sgwc_sess_t *sess = NULL;
+    sgwc_bearer_t *bearer = NULL;
+    sgwc_tunnel_t *tunnel = NULL;
     uint8_t cause_value = 0;
 
     ogs_assert(s11_xact);
@@ -1382,11 +1583,39 @@ void sgwc_s11_handle_delete_indirect_data_forwarding_tunnel_request(
         sgwc_ue->mme_s11_teid, sgwc_ue->sgw_s11_teid);
 
     ogs_list_for_each(&sgwc_ue->sess_list, sess) {
+        bool has_indirect = false;
+        ogs_list_for_each(&sess->bearer_list, bearer) {
+            ogs_list_for_each(&bearer->tunnel_list, tunnel) {
+                if (tunnel->interface_type ==
+                        OGS_GTP2_F_TEID_SGW_GTP_U_FOR_DL_DATA_FORWARDING ||
+                    tunnel->interface_type ==
+                        OGS_GTP2_F_TEID_SGW_GTP_U_FOR_UL_DATA_FORWARDING) {
+                    has_indirect = true;
+                    break;
+                }
+            }
+            if (has_indirect) break;
+        }
 
-        ogs_assert(OGS_OK ==
-            sgwc_pfcp_send_session_modification_request(
-                sess, s11_xact, gtpbuf,
-                OGS_PFCP_MODIFY_INDIRECT| OGS_PFCP_MODIFY_REMOVE));
+        if (has_indirect) {
+            ogs_assert(OGS_OK ==
+                sgwc_pfcp_send_session_modification_request(
+                    sess, s11_xact->id, gtpbuf,
+                    OGS_PFCP_MODIFY_INDIRECT|OGS_PFCP_MODIFY_REMOVE));
+        } else {
+            ogs_error("No Indirect Tunnel");
+            ogs_error("    UE IMSI[%s] APN[%s]",
+                    sgwc_ue->imsi_bcd, sess->session.name);
+            ogs_error("    MME_S11_TEID[%d] SGW_S11_TEID[%d]",
+                    sgwc_ue->mme_s11_teid, sgwc_ue->sgw_s11_teid);
+            ogs_list_for_each(&sess->bearer_list, bearer) {
+                ogs_error("    EBI[%d]", bearer->ebi);
+                ogs_list_for_each(&bearer->tunnel_list, tunnel) {
+                    ogs_error("TUNNEL[%d] INF[%d]",
+                            tunnel->id, tunnel->interface_type);
+                }
+            }
+        }
     }
 }
 
@@ -1472,7 +1701,7 @@ void sgwc_s11_handle_bearer_resource_command(
      * Check ALL Context
      ********************/
     ogs_assert(bearer);
-    sess = bearer->sess;
+    sess = sgwc_sess_find_by_id(bearer->sess_id);
     ogs_assert(sess);
     ogs_assert(sess->gnode);
     ogs_assert(sgwc_ue);
@@ -1492,7 +1721,8 @@ void sgwc_s11_handle_bearer_resource_command(
     }
 
     s5c_xact = ogs_gtp_xact_local_create(
-            sess->gnode, &message->h, pkbuf, gtp_bearer_timeout, bearer);
+            sess->gnode, &message->h, pkbuf, gtp_bearer_timeout,
+            OGS_UINT_TO_POINTER(bearer->id));
     if (!s5c_xact) {
         ogs_error("ogs_gtp_xact_local_create() failed");
         return;
